"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""


# -*- coding: utf-8 -*-
import citrusleaf as cl
from helpers import convertor
from helpers.debugger import Debugger
from helpers.caching import CachedMethod
from constants import *
from helpers.info_dict import InfoDict
from datetime import datetime
from distutils import version
import re

# from flaskapp import app
# logger = app.logger


debugger = Debugger()


@debugger.check_and_log
@CachedMethod(4)
def split_ip_port(node_address):
    match = re.search('(?P<addr>.*):(?P<port>\d+)', node_address)
    ip = match.group('addr')
    port = match.group('port')

    return [ip, port]

@debugger.check_and_log
# @CachedMethod()
def is_num(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

@debugger.check_and_log
# @CachedMethod()
def to_num(s):
    try:
        return int(s)
    except ValueError:
        return float(s)

@debugger.check_and_log
# @CachedMethod()
def get_node_id(ip, port, sock=None):

    return get_node_info(ip, port, sock=sock).get("node", None)

@debugger.check_and_log
def is_node_available(ip, port, sock=None):

    if get_node_info(ip, port, sock=sock) == {}:
        return False
    return True

@debugger.check_and_log
def get_node_build(ip, port, sock=None):

    node_build_id = get_node_info(ip, port, sock=sock).get("build", None)
    return node_build_id

@CachedMethod(4)
def get_cluster_name(ip, port, sock=None, build=None):

    if build is None or version.LooseVersion(build) < version.LooseVersion("3.10.0"):
        return "n/a"
    else:
        name = cl.citrusleaf_info(ip, port, "cluster-name", sock=sock)
        if name == UNABLE_TO_CONNECT or name is None or name == "null":
            return "n/a"
        return name


@debugger.check_and_log
# @CachedMethod(60)
def get_self_address(ip, port, sock=None):

    node_address = cl.citrusleaf_info(ip, port, "service", sock=sock)
    if node_address == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT
    return node_address.split(";")[0]  # multiple_interfaces_on_node

# The following is used to read a segment from peer list, such as a string or number
def read_segment(s, delim=","):
    if not s or len(s) <= 0:
        return["", ""]

    idx = 0
    for ch in s:
        if ch == delim:
            return [s[:idx].strip(), s[idx+1:] if len(s) > idx+1 else ""]

        idx+=1

    return [s.strip(), ""]

# the foloowing breaks down comma delimited lists into arrays of strings
def read_list(s, delim=","):
    if not s or len(s) <= 0 or s[0] != "[":
        return ["", s]

    result = []
    brakets = 0
    segment = ""
    idx = 0
    for ch in s:
        idx += 1
        if ch == delim and brakets == 1:
            result.append(segment.strip())
            segment = ""
            continue
        elif ch == "[":
            brakets += 1
            if brakets == 1:
                continue
        elif ch == "]":
            brakets -= 1
            if brakets == 0:
                if segment != "":
                    result.append(segment.strip())
                segment = ""
                break

        segment += ch


    if segment != "":
        result.append(segment.strip())

    return [result if len(result) > 0 else None, s[idx+1:] if len(s) > idx+1 else ""]

def parse_peers(s):
    gen, s = read_segment(s)
    port, s = read_segment(s)

    nodes = []
    node, _ = read_list(s)
    if not node:
        return []

    for s in node:
        node, _ = read_list(s)
        if node:
            nodes.append(node)

    results = []
    for i, node in enumerate(nodes):
        addresses = node[2]
        addresses = read_list(addresses)[0]

        # add default port to the addresses which do not have port
        for a in addresses:
            if a and len(a) > 0:
                if a[0] == "[":
                    if a[-1] == "]":
                        # ipv6 without port
                        results.append(a[1:-1] + ":" + port)
                    else:
                        a = a.replace("[", "")
                        a = a.replace("]", "")
                        results.append(a)

                    continue
                elif ":" not in a:
                        # ipv4 or dns without port
                        results.append(a + ":" + port)
                        continue
                results.append(a)

    return results

@debugger.check_and_log
def get_cluster_nodes(ip, port, sock=None, build=None):

    if build is None or version.LooseVersion(build) < version.LooseVersion("3.10.0"):
        services = get_node_info(ip, port, sock=sock).get("services", UNABLE_TO_CONNECT)
        service = get_self_address(ip, port, sock=sock)
        nodes = []
        if services == UNABLE_TO_CONNECT or service == UNABLE_TO_CONNECT:
            return nodes
        nodes = convertor.text_to_list(services)
        nodes.append(service.split(";")[0])  # multiple_interfaces_on_node
    else:
        services = cl.citrusleaf_info(ip, port, 'peers-clear-std', sock=sock)
        service = get_self_address(ip, port, sock=sock)
        if services == UNABLE_TO_CONNECT or service == UNABLE_TO_CONNECT:
            return []

        nodes = parse_peers(services)
        nodes.append(service.split(";")[0])

    return nodes


@debugger.check_and_log
def get_cluster_nodes_alternate(ip, port, sock=None, build=None):
    """ Get alternate-address of nodes"""
    node_list = list()

    if build is None or version.LooseVersion(build) < version.LooseVersion("3.10.0"):
        # citrusleaf_info() throws Attribute Error for some AS versions
        try:
            cluster_nodes = cl.citrusleaf_info(
                ip, port, 'services-alternate', sock=sock)  # Returns external IP
        except AttributeError:
            return node_list

        if cluster_nodes != UNABLE_TO_CONNECT and \
                cluster_nodes not in ERROR_DESCRIPTION and cluster_nodes:
            node_list = convertor.text_to_list(cluster_nodes)

        return node_list
    else:
        try:
            cluster_nodes = cl.citrusleaf_info(
                ip, port, 'peers-clear-alt', sock=sock)  # Returns external IP
        except AttributeError:
            return node_list

        if cluster_nodes != UNABLE_TO_CONNECT and \
                cluster_nodes not in ERROR_DESCRIPTION and cluster_nodes:
            node_list = parse_peers(cluster_nodes)

        return node_list


@debugger.check_and_log
def get_node_statistics(ip, port, sock=None):

    statistics = get_node_info(ip, port, sock=sock).get("statistics", None)
    if statistics is None:
        return InfoDict(node_status="off")
    statistics = convertor.text_to_list(statistics)
    statistics = convertor.list_to_dict(statistics)
    statistics['timestamp'] = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
    return statistics

@debugger.check_and_log
# @CachedMethod(300)
def get_node_info(ip, port, sock=None):

    info = cl.citrusleaf_info(ip, port, sock=sock)
    if info == UNABLE_TO_CONNECT:
        return {}
    return info

@debugger.check_and_log
# @CachedMethod()
def get_node_configuration(ip, port, sock=None):

    configuration = cl.citrusleaf_info(ip, port, "get-config:", sock=sock)
    if configuration == UNABLE_TO_CONNECT:
        return InfoDict(node_status="off")
    configuration = convertor.text_to_list(configuration)
    return convertor.list_to_dict(configuration)

@debugger.check_and_log
# @CachedMethod()
def get_node_bins_info(ip, port, sock=None):

    bins_info = dict()
    temp = cl.citrusleaf_info(ip, port, "bins", sock=sock)
    if temp == UNABLE_TO_CONNECT:
        return None

    for row in convertor.text_to_list(temp):
        if row:
            namespace, bin_info = convertor.text_to_list(row, ":")
            temp = convertor.text_to_list(bin_info, ",")
            bins_info[namespace] = convertor.list_to_dict(temp[:2])
            bins_info[namespace]["bins"] = temp[2:]

    return bins_info

@debugger.check_and_log
# @CachedMethod()
def get_namespaces(ip, port, sock=None):

    namespaces = cl.citrusleaf_info(ip, port, "namespaces", sock=sock)
    if namespaces == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT
    return convertor.text_to_set(namespaces)

@debugger.check_and_log
# @CachedMethod()
def get_agg_namespace_statistics(ip, port, sock=None):

    agg_stats = InfoDict()
    ns_list = get_namespaces(ip, port, sock=sock)
    if ns_list == UNABLE_TO_CONNECT:
        return None
    for ns in ns_list:
        stats = get_namespace_statistics(ns, ip, port, sock=sock)
        for stat, val in stats.iteritems():
            if is_num(val):
                agg_stats[stat] = agg_stats.get(stat, 0) + to_num(val)

    for stat, val in agg_stats.iteritems():
        agg_stats[stat] = str(val)

    return agg_stats

@debugger.check_and_log
# @CachedMethod()
def get_namespace_statistics(namespace, ip, port, sock=None):

    stats = cl.citrusleaf_info(ip, port, "namespace/%s" % namespace, sock=sock)
    if stats == UNABLE_TO_CONNECT:
        return InfoDict(node_status="off")
    stats = convertor.text_to_list(stats)
    return convertor.list_to_dict(stats)

@debugger.check_and_log
# @CachedMethod()
def get_sets_info(ip, port, sock=None):

    sets_info = dict()
    result = cl.citrusleaf_info(ip, port, "sets", sock=sock)

    if result == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT

    if result:
        for row in convertor.text_to_list(result.strip(), ";"):
            if not row:
                continue

            details = convertor.list_to_dict(
                convertor.text_to_list(row, ":"), "=")
            for key in list(details.keys()):  # For Aerospike 3.0.22+
                if key.startswith('set-'):
                    new_key = key.replace('set-', '')
                    details[new_key] = details.pop(key)

            ns = 'ns_name'
            if details.get('ns_name') is None:
                ns = 'ns'
                # alias the keys
                details['ns_name'] = details.get('ns')
                details['set_name'] = details.get('set')
                details['n_objects'] = details.get('objects')
                details['stop-write-count'] = details.get('stop-writes-count')
                details['delete'] = details.get('deleting')
                details['evict-hwm-count'] = 'n/s'

            if details.get(ns) is not None:
                namespace = sets_info.get(details[ns], [])
                namespace.append(details)
                sets_info[details[ns]] = namespace

    return sets_info


@debugger.check_and_log
# @CachedMethod()
def get_dc_statistics(ip, port, dc, sock=None):

    statistics = cl.citrusleaf_info(ip, port, "dc/%s" % dc, sock=sock)
    if statistics == UNABLE_TO_CONNECT:
        return InfoDict()
    statistics = convertor.text_to_list(statistics)
    return convertor.list_to_dict(statistics)

@debugger.check_and_log
# @CachedMethod()
def get_dc_names(ip, port, sock=None):

    names = cl.citrusleaf_info(ip, port, "dcs", sock=sock)
    if names == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT

    return convertor.text_to_list(names)

@debugger.check_and_log
# @CachedMethod()
def get_udf_list(ip, port, sock=None):

    lst = cl.citrusleaf_info(ip, port, "udf-list", sock=sock)
    if lst == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT

    udfs = []
    for record in convertor.text_to_list(lst, ";"):
        if record != '':
            udfs.append(convertor.list_to_dict(
                convertor.text_to_list(record, ",")))

    return udfs

@debugger.check_and_log
# @CachedMethod()
def get_scan_jobs(ip, port, sock=None):

    lst = cl.citrusleaf_info(ip, port, "scan-list", sock=sock)
    if lst == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT
    elif lst == "No running scans":
        return []

    jobs = []
    for record in convertor.text_to_list(lst):
        jobs.append(convertor.list_to_dict(convertor.text_to_list(record, ":")))

    return jobs

@debugger.check_and_log
# @CachedMethod()
def get_job_details1(ip, port, sock=None):

    lst = cl.citrusleaf_info(ip, port, "mon:cmd=getall-job", sock=sock)
    if lst == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT
    elif lst == "No running scans" or lst == "" or lst == "error":
        return []

    jobs = {}
    keys = ["trid", "cpu", "mem", "start_time", "run_time", "disk_io_bytes", "net_io_bytes", "priority"]
    for record in lst[1:-1].split("]["):
        details = convertor.text_to_list(record, "|")
        details = dict(zip(keys, details))
        details["address"] = "%s:%s" % (ip, port)
        jobs[details["trid"]] = details

    return jobs

@debugger.check_and_log
# @CachedMethod()
def get_job_details(ip, port, sock=None):

    lst = cl.citrusleaf_info(ip, port, "jobs:", sock=sock)
    if lst == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT
    elif lst == "No running scans" or lst == "" or lst == "error":
        return []

    jobs = {}
    for record in convertor.text_to_list(lst.strip(), ";"):
        details = convertor.list_to_dict(
            convertor.text_to_list(record, ":"), "=")
        details["address"] = "%s:%s" % (ip, port)
        jobs[details["trid"]] = details

    return jobs

@debugger.check_and_log
# @CachedMethod()
def get_index_stats(ip, port, namespace, index, sock=None):

    command = "sindex/%s/%s" % (str(namespace), str(index))
    stats = cl.citrusleaf_info(ip, port, command, sock=sock)
    if stats == UNABLE_TO_CONNECT:
        return InfoDict(node_status="off")
    elif stats == "ns_type=unknown":
        return InfoDict(namespace_found=False)
    elif stats == "Invalid StatsFAIL:201:NO INDEX":
        return InfoDict(index_found=False)
    stats = convertor.text_to_list(stats)
    return convertor.list_to_dict(stats)

@debugger.check_and_log
# @CachedMethod()
def get_indexes(ip, port, sock=None):

    output = cl.citrusleaf_info(ip, port, "sindex", sock=sock)
    if output == UNABLE_TO_CONNECT:
        return UNABLE_TO_CONNECT

    indexes = {}
    output = output.replace("sync_state=:", "sync_state=")
    for record in convertor.text_to_list(output):
        details = convertor.list_to_dict(convertor.text_to_list(record, ":"))
        if "ns" not in details:
            continue
        namespace = indexes.get(details["ns"], [])
        namespace.append(details)
        indexes[details["ns"]] = namespace

    return indexes
